/*
 * File: html.c
 *
 * Copyright (C) 1997 Raph Levien <raph@acm.org>
 * Copyright (C) 1999 James McCollough <jamesm@gtwn.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

/*
 * Dillo HTML parsing routines
 */

#include <ctype.h>              /* for isspace and tolower */
#include <string.h>             /* for memcpy and memmove */
#include <stdlib.h>             /* for atoi */
#include <stdio.h>              /* for sprintf */

#include <gtk/gtk.h>

#include "list.h"
#include "colors.h"
#include "html.h"
#include "web.h"
#include "IO/Url.h"
#include "dillo.h"
#include "nav.h"
#include "browser.h"
#include "dicache.h"

#include "dw.h"
#include "dw_page.h"
#include "dw_bullet.h"
#include "dw_hruler.h"
#include "dw_embedgtk.h"
#include "IO/IO.h"
#include "cache.h"
#include "interface.h"
#include "progressbar.h"
#include "prefs.h"

typedef void (*TagFunct) (DilloHtml *Html, char *Tag, gint Tagsize);


/*
 * Forward declarations
 */
static gint Html_get_attr(const char *tag, gint tagsize,
                           const char *param, char *data, gint datasize);
static void Html_write(DilloHtml *html, char *Buf, gint BufSize);
static void Html_close(DilloHtml *html);
static void Html_callback(int Op, void *CbData, guchar *Buf, gint BufSize);
static DilloHtml *Html_new(BrowserWindow *bw);

/*
 * Local Data
 */
/* -RL :: I use this var for <OL><\OL> tags */
static gint number;


/*
 * Set callback function and callback data for "html/text" MIME type.
 */
Dw *a_Html_text(const char *Type, void *P, CA_Callback_t *Call, void **Data)
{
   DilloWeb *web = P;
   DilloHtml *html = Html_new(web->bw);

   *Data = (void *) html;
   *Call = (CA_Callback_t) Html_callback;

   /* todo: this should probably be a function of dw_page */
   html->linkblock->bw = web->bw;
   web->linkblock = html->linkblock;

   return html->dw;
}


/* 
 * We'll make the linkblock first to get it out of the way.
 */
static DilloHtmlLB *Html_lb_new(void)
{
   DilloHtmlLB *html_lb = g_new(DilloHtmlLB, 1); 

   html_lb->bw = NULL;
   html_lb->base_url = NULL;
   html_lb->num_forms_max = 2;
   html_lb->num_forms = 0;
   html_lb->forms = NULL; 
   html_lb->NRef = 0;
   return html_lb;
}

/*
 * ?
 */
static void Html_lb_free(void *D)
{
   gint i, j;
   DilloHtmlForm *form;
   DilloHtmlLB *html_lb = D;

   if (html_lb->NRef) {
      html_lb->NRef--;
      return;
   }
   for (i = 0; i < html_lb->num_forms; i++) {
      form = &html_lb->forms[i];
      g_free(form->action);
      for (j = 0; j < form->num_inputs; j++) {
         g_free(form->inputs[j].name);
         g_free(form->inputs[j].init_str);
         /* todo: free junk in the select. */
      }
   }
   g_free(html_lb->forms);
   if (html_lb->base_url)
      g_free(html_lb->base_url);
   html_lb->base_url = NULL;
   g_free(html_lb);
}

/* 
 * Handle the status function generated by the dw scroller,
 * and show the url in the browser status-bar.
 */
static void Html_handle_status(void *data, const char *url)
{
   a_Interface_status(data, "%s", url);
}

/*
 * Allocate and insert form information into the Html linkblock
 */
static gint Html_form_new(DilloHtmlLB *html_lb,
                          DilloHtmlMethod method,
                          const char *action,
                          DilloHtmlEnc enc)
{
   gint nf;

   a_List_add(html_lb->forms, html_lb->num_forms, sizeof(*html_lb->forms),
              html_lb->num_forms_max);

   nf = html_lb->num_forms;
   html_lb->forms[nf].method = method;
   html_lb->forms[nf].action = g_strdup(action);
   html_lb->forms[nf].enc = enc;
   html_lb->forms[nf].num_inputs = 0;
   html_lb->forms[nf].num_inputs_max = 4;
   html_lb->forms[nf].inputs = NULL;
   html_lb->forms[nf].HasSubmitButton = FALSE;
   html_lb->num_forms++;

   g_print("Html_form_new: action=%s nform=%d\n", action, nf);
   return nf;
}

/*
 * ?
 */
static void Html_page_check(DilloHtml *html)
{
char *PageFonts[] = { "helvetica" , "lucida", "courier", "utopia" };

   Dw *page;
   DwPageFont font;
   DwPageAttr attr;

   if (html->dw)
      return;

   page = a_Dw_page_new(&html->dw);
   html->dw = page;

   /* Create a dummy font, attribute, and tag for the bottom of the stack. */
#if 0
   font.font = NULL;
#endif
   font.name = PageFonts[0]; /* Helvetica */
   font.size = 12;
   font.bold = FALSE;
   font.italic = FALSE;

   a_Dw_page_init_attr((DwPage *) page, &attr);
   attr.font = a_Dw_page_find_font((DwPage *) page, &font);
   html->stack[0].attr = a_Dw_page_find_attr((DwPage *)page, &attr);
   /* Handle it when the user clicks on a link */
   a_Dw_page_set_callbacks((DwPage *) page,
                           a_Nav_vpush, (void *) html->bw,
                           Html_handle_status, html->bw,
                           Html_lb_free, html->linkblock);
}

/*
 * ?
 */
static DilloHtml *Html_new(BrowserWindow *bw)
{
   DilloHtml *html;

   html = g_new(DilloHtml, 1);

   html->Start_Ofs = 0;
   html->dw = NULL;
   html->bw = bw;
   html->linkblock = Html_lb_new();

   html->stack_max = 16;
   html->stack_top = 0;
   html->stack = g_new(DilloHtmlState, html->stack_max);
   html->stack[0].tag = g_strdup("dillo");
   html->stack[0].parse_mode = DILLO_HTML_PARSE_MODE_INIT;
   html->stack[0].list_level = 0;
   html->stack[0].form = 0;
/*
                         Html_form_new(html->linkblock,
                                       DILLO_HTML_METHOD_GET,
                                       "pripri:",
                                       DILLO_HTML_ENC_URLENCODING);
*/

   html->stash_size_max = 1024;
   html->stash_size = 0;
   html->stash_buf = g_new(char, html->stash_size_max);

   Html_page_check(html);
   return html;
}

/*
 * Initialize the stash buffer 
 */
static void Html_stash_init(DilloHtml *html)
{
   html->stack[html->stack_top].parse_mode = DILLO_HTML_PARSE_MODE_STASH;
   html->stash_size = 0;
   html->stash_space = 0;
}

/*
 * Convert the entities in a token to plain ISO character codes. Takes
 * a token in buf, bufsize format and returns a newly allocated
 * string. 
 * (it takes a lot of time, according to profiling)
 */
static char *Html_parse_entities(char *token, gint num)
{
   char *new_str;
   gint in_index, out_index;

   /* BUG: this should be a perfect hash table */
   struct {
      char *entity;
      gint isocode;
   } entities[] = {
      { "quot", 34 },
      { "amp", 38 },
      { "lt", 60 },
      { "gt", 62 },
      { "nbsp", 32 },
      { "Aacute", 0301 },
      { "aacute", 0341 },
      { "Acirc", 0302 },
      { "acirc", 0342 },
      { "AElig", 0306 },
      { "aelig", 0346 },
      { "Agrave", 0300 },
      { "agrave", 0340 },
      { "Aring", 0305 },
      { "aring", 0345 },
      { "Atilde", 0303 },
      { "atilde", 0343 },
      { "Auml", 0304 },
      { "auml", 0344 },
      { "Ccedil", 0307 },
      { "ccedil", 0347 },
      { "Eacute", 0311 },
      { "eacute", 0351 },
      { "Ecirc", 0312 },
      { "ecirc", 0352 },
      { "Egrave", 0310 },
      { "egrave", 0350 },
      { "ETH", 0320 },
      { "eth", 0360 },
      { "Euml", 0313 },
      { "euml", 0353 },
      { "Iacute", 0315 },
      { "iacute", 0355 },
      { "Icirc", 0316 },
      { "icirc", 0356 },
      { "Igrave", 0314 },
      { "igrave", 0354 },
      { "Iuml", 0317 },
      { "iuml", 0357 },
      { "Ntilde", 0321 },
      { "ntilde", 0361 },
      { "Oacute", 0323 },
      { "oacute", 0363 },
      { "Ocirc", 0324 },
      { "ocirc", 0364 },
      { "Ograve", 0322 },
      { "ograve", 0362 },
      { "Oslash", 0330 },
      { "oslash", 0370 },
      { "Otilde", 0325 },
      { "otilde", 0365 },
      { "Ouml", 0326 },
      { "ouml", 0366 },
      { "szlig", 0337 },
      { "THORN", 0336 },
      { "thorn", 0376 },
      { "Uacute", 0332 },
      { "uacute", 0372 },
      { "Ucirc", 0333 },
      { "ucirc", 0373 },
      { "Ugrave", 0331 },
      { "ugrave", 0371 },
      { "Uuml", 0334 },
      { "uuml", 0374 },
      { "Yacute", 0335 },
      { "yacute", 0375 },
      { "yuml", 0377 },
      { "reg", 174 },
      { "copy", 169 }
      /* There are also the "spyglass" entities, but I think these will do. */
   };
   gint isocode;
   gint i;
   gint n = sizeof(entities) / sizeof(entities[0]);
   gint len;

   new_str = g_new(char, num + 1);

   out_index = 0;
   for (in_index = 0; in_index < num; in_index++) {
      if (token[in_index] == '&') {

         /* it's an entity */
         in_index++;

         if (in_index < num && token[in_index] == '#') {
            /* numeric token -- THIS IS TRICKY, CAUSE THE STRING MAY NOT BE
             * TERMINATED! */
            isocode = atoi(token + in_index + 1);
            if (isocode != 0)
               new_str[out_index++] = isocode;
         } else {
            /* search for named entity */
            for (i = 0; i < n; i++) {
               len = strlen(entities[i].entity);
               if (in_index + len + 1 <= num &&
                   token[in_index + len] == ';' &&
                   !memcmp(token + in_index, entities[i].entity, len)) {
                  new_str[out_index++] = entities[i].isocode;
                  break;
               }
            }
         }

         /* skip past closing semicolon */
         {
            void *Ptr = memchr(token + in_index, ';', num - in_index);

            if (Ptr)
               in_index = (gulong) Ptr - (gulong) token;
         }
      } else {
         new_str[out_index++] = token[in_index];
      }
   }
   new_str[out_index] = '\0';
   return new_str;
}

/*
 * ?
 */
static void Html_process_word(DilloHtml *html, char *word, gint size)
{
   if (html->stack[html->stack_top].parse_mode == DILLO_HTML_PARSE_MODE_STASH){
      if (html->stash_size + size + 2 > html->stash_size_max) {
         do {
            html->stash_size_max <<= 1;
         } while (html->stash_size + size + 2 > html->stash_size_max);
         html->stash_buf = g_realloc(html->stash_buf, html->stash_size_max);
      }
      if (html->stash_space)
         html->stash_buf[html->stash_size++] = ' ';
      html->stash_space = FALSE;
      memcpy(html->stash_buf + html->stash_size, word, size);
      html->stash_size += size;
      html->stash_buf[html->stash_size] = '\0';
   } else {
      if (word[0] == '\n')
         a_Dw_page_linebreak((DwPage *) html->dw);

      /* this works because of the same side-effect that causes the
       * problem.  newlines (and tabs) passed in the string are
       * basically ignored, no need to change the size and word
       * pointer.
       *
       * right now, this support for <pre> tags ignore tabs entirely.
       * I can't find a working way to insert more than one space
       * (i.e., a tab).. when I tried inserting a string of spaces,
       * the program would seg fault when it tried to follow a link.
       * Using gtk_page_add_space multiple times also failed. -Otto
       */

      a_Dw_page_add_text((DwPage *) html->dw,
                         Html_parse_entities(word, size),
                         html->stack[html->stack_top].attr);
   }
}

/*
 * ?
 */
static void Html_process_space(DilloHtml *html, char *space, gint spacesize)
{
   if (html->stack[html->stack_top].parse_mode == DILLO_HTML_PARSE_MODE_STASH){
      html->stash_space = (html->stash_size > 0);
   } else {
      a_Dw_page_add_space((DwPage *) html->dw,
                          html->stack[html->stack_top].attr);
   }
}

/*
 * Does the tag in tagstr (e.g. "p") match the tag in the tag, tagsize
 * structure, with the initial < skipped over (e.g. "P align=center>")
 */
static gboolean Html_match_tag(const char *tagstr, char *tag, gint tagsize)
{
   gint i;

   for (i = 0; i < tagsize && tagstr[i] != '\0'; i++) {
      if (tolower(tagstr[i]) != tolower(tag[i]))
         return FALSE;
   }
   /* The test for '/' is for xml compatibility: "empty/>" will be matched. */
   if (i < tagsize && (isspace(tag[i]) || tag[i] == '>' || tag[i] == '/'))
      return TRUE;
   return FALSE;
}

/*
 * Push the tag (copying attributes from the top of the stack) 
 */
static void Html_push_tag(DilloHtml *html, char *tag, gint tagsize)
{
   char *tagstr;
   gint tag_end;
   gint n_items;

   /* Copy the tag itself (no parameters) into tagstr. */
   tag_end = 1;
   while (tag_end < tagsize && isalnum(tag[tag_end]))
      tag_end++;
   tagstr = g_new(char, tag_end);
   memcpy(tagstr, tag + 1, tag_end - 1);
   tagstr[tag_end - 1] = 0;

   n_items = html->stack_top + 1;
   a_List_add(html->stack, n_items, sizeof(*html->stack), html->stack_max);
   /* We'll copy the former stack item and just change the tag
    * instead of copying all fields except for tag.  --Jcid */
   html->stack[n_items] = html->stack[n_items - 1];
   html->stack[n_items].tag = tagstr;
   html->stack_top = n_items;
}

/*
 * ?
 */
static void Html_cleanup_tag(DilloHtml *html, char *tag)
{
   if ( html->stack_top &&
        Html_match_tag(html->stack[html->stack_top].tag, tag, strlen(tag)) )
      g_free(html->stack[html->stack_top--].tag);
}

/*
 * Do a paragraph break and push the tag. This pops unclosed <p> tags
 * off the stack, if there are any. 
 */
static void Html_par_push_tag(DilloHtml *html, char *tag, gint tagsize)
{
   Html_cleanup_tag(html, "p>");
   Html_push_tag(html, tag, tagsize);
   a_Dw_page_parbreak((DwPage *) html->dw, 9);
}

/* 
 * Pop the tag. At the moment, this assumes tags nest! 
 */
static void Html_pop_tag(DilloHtml *html, char *tag, gint tagsize)
{
   gint tag_index;

   /* todo: handle non-nesting case more gracefully (search stack
    * downwards for matching tag). */

   for (tag_index = html->stack_top; tag_index > 0; tag_index--) {
      if (Html_match_tag(html->stack[tag_index].tag, tag + 2, tagsize - 2)) {
         while (html->stack_top >= tag_index)
            g_free(html->stack[html->stack_top--].tag);
         return;
      }
   }
   /* Not found, just ignore. */
}

/*
 * ?
 */
static void Html_tag_open_title(DilloHtml *html, char *tag, gint tagsize)
{

   Html_push_tag(html, tag, tagsize);
   Html_stash_init(html);
}

/*
 * ?
 */
static void Html_tag_close_title(DilloHtml *html, char *tag, gint tagsize)
{

#ifdef VERBOSE
   g_print("Html_tag_close_title: %s\n", html->stash_buf);
#endif
   a_Interface_set_Page_title(html->linkblock->bw, html->stash_buf);
   Html_pop_tag(html, tag, tagsize);
}

/*
 * ?
 */
static void Html_tag_open_body(DilloHtml *html, char *tag, gint tagsize)
{
   char bodycolor[16];
   DwPage *page;
   DwPageAttr attr;
   gint32 color;
   guint idx;

   page = (DwPage *)html->dw;

   if (Html_get_attr(tag, tagsize, "bgcolor", bodycolor, sizeof(bodycolor))){
      color = a_Color_parse (bodycolor, prefs.bg_color);
      if ( (color == 0xffffff && !prefs.allow_white_bg) 
           || prefs.force_my_colors )
         color = prefs.bg_color;
      a_Dw_set_color(&page->dw, color);

      /* We set 'page->bgnd_color' out of a_Dw_set_color, because that funtion
       * is meant to be general, not just for DwPage  --Jcid */
      page->bgnd_color = color;
   }   
   
   if (Html_get_attr (tag, tagsize, "text", bodycolor, sizeof(bodycolor))) {
      color = a_Color_parse (bodycolor, prefs.text_color);
      if (prefs.force_my_colors) color = prefs.text_color;
      idx = html->stack[html->stack_top].attr;
      attr = page->attrs[idx];
      attr.color = a_Dw_page_find_color(page, color);
      html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
   }

   if (Html_get_attr(tag, tagsize, "link", bodycolor, sizeof(bodycolor))) {
      if (prefs.force_my_colors) 
         page->link_color = prefs.link_color;
      else
         page->link_color = a_Color_parse (bodycolor, prefs.link_color);
   }
 
   Html_push_tag (html, tag, tagsize);
   html->stack[html->stack_top].parse_mode = DILLO_HTML_PARSE_MODE_BODY;
}

/*
 * ?
 */
static void Html_tag_open_p(DilloHtml *html, char *tag, gint tagsize)
{
   Html_par_push_tag(html, tag, tagsize);
}

/*
 * This function is just for not confusing the parser after tables,
 * the real TABLE implementation will start soon... (Hopefully)
 */
static void Html_tag_open_td(DilloHtml *html, char *tag, gint tagsize)
{
   Html_push_tag(html, tag, tagsize);
}

/*
 * ?
 */
static void Html_tag_open_h(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageFont font;
   DwPageAttr attr;

   gint level;
   gint sizes[] =
   {24, 18, 18, 14, 12, 10};

   Html_par_push_tag(html, tag, tagsize);
   level = tag[2] - '0';
   level = MAX(1, level);
   level = MIN(6, level);

   page = (DwPage *) html->dw;
   attr = page->attrs[html->stack[html->stack_top].attr];
   font = page->fonts[attr.font];
   font.size = sizes[level - 1];
   font.bold = TRUE;
   font.name = "helvetica";
   attr.font = a_Dw_page_find_font(page, &font);
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
}

/*
 * ?
 */
static void Html_tag_open_br(DilloHtml *html, char *tag, gint tagsize)
{
   a_Dw_page_linebreak((DwPage *) html->dw);
}

/*
 * ?
 */
static void Html_tag_open_b(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageFont font;
   DwPageAttr attr;

   Html_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;
   attr = page->attrs[html->stack[html->stack_top].attr];
   font = page->fonts[attr.font];
   font.bold = TRUE;
   attr.font = a_Dw_page_find_font(page, &font);
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
}

/*
 * ?
 */
static void Html_tag_open_strong(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageFont font;
   DwPageAttr attr;

   Html_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;
   attr = page->attrs[html->stack[html->stack_top].attr];
   font = page->fonts[attr.font];
   font.bold = TRUE;
   attr.font = a_Dw_page_find_font(page, &font);
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
}

/*
 * ?
 */
static void Html_tag_open_i(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageFont font;
   DwPageAttr attr;

   Html_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;
   attr = page->attrs[html->stack[html->stack_top].attr];
   font = page->fonts[attr.font];
   font.italic = TRUE;
   attr.font = a_Dw_page_find_font(page, &font);
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
}

/*
 * ?
 */
static void Html_tag_open_em(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageFont font;
   DwPageAttr attr;

   Html_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;
   attr = page->attrs[html->stack[html->stack_top].attr];
   font = page->fonts[attr.font];
   font.italic = TRUE;
   attr.font = a_Dw_page_find_font(page, &font);
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
}

/*
 * ?
 */
static void Html_tag_open_cite(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageFont font;
   DwPageAttr attr;

   Html_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;
   attr = page->attrs[html->stack[html->stack_top].attr];
   font = page->fonts[attr.font];
   font.italic = TRUE;
   attr.font = a_Dw_page_find_font(page, &font);
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
}

/*
 * ?
 */
static void Html_tag_open_tt(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageFont font;
   DwPageAttr attr;

   Html_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;
   attr = page->attrs[html->stack[html->stack_top].attr];
   font = page->fonts[attr.font];
   font.name = "courier";
   attr.font = a_Dw_page_find_font(page, &font);
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
}

/* todo: remove this - it's not standard html */
/*
 * ?
 */
static void Html_tag_open_button(DilloHtml *html, char *tag, gint tagsize)
{
   /* don't bother */
}

/*
 * Create a new imgsink, and request the image (url) to the cache
 * (If it either hits or misses, is not relevant here; that's up to the
 *  cache functions)
 */
static void Html_tag_open_img(DilloHtml *html, char *tag, gint tagsize)
{
   DilloImgSink *imgsink;
   char src[1024];
   gchar *url;
   char num[64];
   char alt[1024];
   gint width, height;
   char *alt_ptr;
   Dw *dw;
   DwPage *page = (DwPage *)html->dw;

   if ( !Html_get_attr(tag, tagsize, "src", src, sizeof(src)) )
      return;
   url = a_Url_resolve_relative(html->linkblock->base_url, src);
   if ( !url )
      return;

   width = 0;
   if (Html_get_attr(tag, tagsize, "width", num, sizeof(num)))
      width = atoi(num);
   height = 0;
   if (Html_get_attr(tag, tagsize, "height", num, sizeof(num)))
      height = atoi(num);
   alt_ptr = NULL;
   if (Html_get_attr(tag, tagsize, "alt", alt, sizeof(alt)))
      alt_ptr = alt;

   imgsink = a_Image_new(0, 0, alt_ptr, page->bgnd_color);
   dw = (Dw *) imgsink->dw;

   /* note: a_Dicache_request_url_img comes before a_Dw_page_add_widget
    * because cached images set image size, avoiding a resize later  --Jcid */
   html->linkblock->NRef++;
   a_Dicache_request_url_img(imgsink, url, html->linkblock->bw);
   /* Dereference the link block */
   Html_lb_free(html->linkblock);

   a_Dw_page_add_widget((DwPage *) html->dw, dw,
                        html->stack[html->stack_top].attr);
   g_free(url);
}

/*
 * ?
 */
static void Html_tag_open_a(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageAttr attr;

   char href[1024];
   gchar *url;
   guint idx;

   Html_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;

   if ( !Html_get_attr(tag, tagsize, "href", href, sizeof(href)) )
      return;
   url = a_Url_resolve_relative(html->linkblock->base_url, href);
   if ( !url )
      return;
   idx = html->stack[html->stack_top].attr;
   attr = page->attrs[idx];
   attr.link = a_Dw_page_new_link(page, url);
   attr.color = a_Dw_page_find_color(page, page->link_color);
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
   g_free(url);
}

/*
 * ?
 */
static void Html_tag_open_blockquote(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageAttr attr;

   Html_par_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;

   attr = page->attrs[html->stack[html->stack_top].attr];
   attr.left_indent_first = attr.left_indent_rest + 40;
   attr.left_indent_rest += 40;
   attr.right_indent += 40;
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
}

/*
 * Handle the <UL> tag.
 */
static void Html_tag_open_ul(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageAttr attr;

   Html_par_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;

   attr = page->attrs[html->stack[html->stack_top].attr];
   attr.left_indent_first = attr.left_indent_rest + 40;
   attr.left_indent_rest += 40;
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);
   if(html->stack[html->stack_top].list_level < 4)
      html->stack[html->stack_top].list_level++;      

}

/*
 * Handle the <OL> tag.
 */
static void Html_tag_open_ol (DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageAttr attr;

   Html_par_push_tag (html, tag, tagsize);

   page = (DwPage *)html->dw;

   attr = page->attrs[html->stack[html->stack_top].attr];
   attr.left_indent_first = attr.left_indent_rest + 40;
   attr.left_indent_rest += 40;

   html->stack[html->stack_top].attr = a_Dw_page_find_attr (page, &attr);
   html->stack[html->stack_top].list_level = 4;  /* NUMBERED */

   number = 0;
}

/*
 * Handle the <LI> tag.
 */
static void Html_tag_open_li (DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageAttr attr;
   Dw *bullet;
   gint indent;
   char *str;

   Html_cleanup_tag(html, "p>");
   Html_cleanup_tag(html, "li>");
   Html_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;

   a_Dw_page_parbreak(page, 1);

   attr = page->attrs[html->stack[html->stack_top].attr];
   indent = attr.left_indent_rest;
   if (indent < 40)
      indent = 40;
   attr.left_indent_first = indent - 12;
   attr.left_indent_rest = indent;

   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);

   switch (html->stack[html->stack_top].list_level) {
   case 1:
      bullet = a_Dw_bullet_new(DW_BULLET_DISC);
      a_Dw_page_add_widget(page, bullet,
                           html->stack[html->stack_top].attr);
      a_Dw_page_add_space(page, html->stack[html->stack_top].attr);
      break;
   case 2:
      bullet = a_Dw_bullet_new(DW_BULLET_CIRCLE);
      a_Dw_page_add_widget(page, bullet, html->stack[html->stack_top].attr);
      a_Dw_page_add_space(page, html->stack[html->stack_top].attr);      
      break;
   case 3:
      bullet = a_Dw_bullet_new(DW_BULLET_SQUARE);
      a_Dw_page_add_widget(page, bullet, html->stack[html->stack_top].attr);
      a_Dw_page_add_space(page, html->stack[html->stack_top].attr);
      break;
   case 4:
      str = g_new(char, 5);
      sprintf(str, "%d. ", ++number);
      a_Dw_page_add_text ((DwPage *) html->dw, str,
                          html->stack[html->stack_top].attr);
      break;
   default:
      /* Never Reached */
      break;  
   }
}

/*
 * Create an horizontal-ruler widget
 */
static void Html_tag_open_hr(DilloHtml *html, char *tag, gint tagsize)
{
   gint width;
   Dw *dw;

   width = ((DwPage *) html->dw)->width;
   dw = a_Dw_hruler_new(width);
   a_Dw_page_add_widget((DwPage *) html->dw, dw,
                        html->stack[html->stack_top].attr);
   a_Dw_page_linebreak((DwPage *) html->dw);
}

/*
 * ?
 */
static void Html_tag_open_dl(DilloHtml *html, char *tag, gint tagsize)
{
   /* may want to actually do some stuff here. */
   Html_par_push_tag(html, tag, tagsize);
} 

/*
 * ?
 */
static void Html_tag_open_dt(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageFont font;
   DwPageAttr attr;

   Html_cleanup_tag(html, "p>");
   Html_cleanup_tag(html, "dd>");
   Html_cleanup_tag(html, "dt>");

   Html_par_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;
   attr = page->attrs[html->stack[html->stack_top].attr];
   font = page->fonts[attr.font];
   font.bold = TRUE;
   attr.font = a_Dw_page_find_font(page, &font);
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);

   /* I really want less space between this and the description, but
    * for some reason using gtk_page_linebreak doesn't work, so I've
    * made the "close tag" function Html_par_close which, of
    * course, gives me an awful lot of vertical space after a <dt>.
    * Purely aesethic, I guess... it'd just be nice to know why.
    */
}

/*
 * ?
 */
static void Html_tag_open_dd(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageAttr attr;

   Html_cleanup_tag(html, "p>");
   Html_cleanup_tag(html, "dd>");
   Html_cleanup_tag(html, "dt>");

   Html_par_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;

   attr = page->attrs[html->stack[html->stack_top].attr];
   attr.left_indent_first = attr.left_indent_rest + 40;
   attr.left_indent_rest += 40;
   attr.right_indent += 40;
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);

}

/*
 * ?
 */
static void Html_tag_open_pre(DilloHtml *html, char *tag, gint tagsize)
{
   DwPage *page;
   DwPageFont font;
   DwPageAttr attr;

   Html_par_push_tag(html, tag, tagsize);

   page = (DwPage *) html->dw;
   attr = page->attrs[html->stack[html->stack_top].attr];
   font = page->fonts[attr.font];
   font.name = "courier";
   attr.font = a_Dw_page_find_font(page, &font);
   html->stack[html->stack_top].attr = a_Dw_page_find_attr(page, &attr);

   /* Is the placement of this statement right? */
   html->stack[html->stack_top].parse_mode = DILLO_HTML_PARSE_MODE_PRE;

}

/*
 * Handle <FORM> tag
 */
static void Html_tag_open_form(DilloHtml *html, char *tag, gint tagsize)
{
   char method_str[80];
   char action_str[1024];
   char enc_str[80];
   gchar *action;
   DilloHtmlMethod method;
   DilloHtmlEnc enc;

   Html_par_push_tag(html, tag, tagsize);

   method = DILLO_HTML_METHOD_GET;
   if (Html_get_attr(tag, tagsize, "method", method_str, sizeof(method_str))){
      if (!g_strcasecmp(method_str, "post"))
         method = DILLO_HTML_METHOD_POST;
      /* todo: maybe deal with unknown methods? */
   }
   if (Html_get_attr(tag, tagsize, "action", action_str, sizeof(action_str))) {
      action = a_Url_resolve_relative(html->linkblock->base_url, action_str);
   } else {
      action = g_strdup(html->linkblock->base_url);
   }
   enc = DILLO_HTML_ENC_URLENCODING;
   if ( Html_get_attr(tag, tagsize, "encoding", enc_str, sizeof(enc_str)) ) {
      /* todo: maybe deal with unknown encodings? */
   }
   html->stack[html->stack_top].form =
       Html_form_new(html->linkblock, method, action, enc);
   g_print("Html_tag_open_form stack_top=%d form_new=%d\n",
           html->stack_top, html->stack[html->stack_top].form);
   g_free(action);
}

/*
 * Set the history of the menu to be consistent with the active menuitem. 
 */
static void Html_select_set_history(DilloHtmlInput *input)
{
   gint i;

   for (i = 0; i < input->select->num_options; i++) {
      if (GTK_CHECK_MENU_ITEM(input->select->options[i].menuitem)->active) {
         gtk_option_menu_set_history(GTK_OPTION_MENU(input->widget), i);
         break;
      }
   }
}

/*
 * Reset the input widget to the initial value.
 */
static void Html_reset_input(DilloHtmlInput *input)
{
   gint i;

   switch (input->type) {
   case DILLO_HTML_INPUT_TEXT:
   case DILLO_HTML_INPUT_PASSWORD:
      gtk_entry_set_text(GTK_ENTRY(input->widget), input->init_str);
      break;
   case DILLO_HTML_INPUT_CHECKBOX:
      gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(input->widget),
                                  input->init_val);
      break;
   case DILLO_HTML_INPUT_RADIO:
      gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(input->widget),
                                  input->init_val);
      break;
   case DILLO_HTML_INPUT_SELECT:
      if (input->select != NULL) {
         /* this is in reverse order so that, in case more than one was
          * selected, we get the last one, which is consistent with handling
          * of multiple selected options in the layout code. */
         for (i = input->select->num_options - 1; i >= 0; i--) {
            if (input->select->options[i].init_val) {
               break;
            }
         }
         if (i < 0)
            i = 0;
         gtk_menu_item_activate(GTK_MENU_ITEM(input->select->options[i].menuitem));
         Html_select_set_history(input);
      }
      break;
   case DILLO_HTML_INPUT_SEL_LIST:
      if (input->select != NULL) {
         for (i = 0; i < input->select->num_options; i++) {
            if (input->select->options[i].init_val) {
               if (input->select->options[i].menuitem->state == GTK_STATE_NORMAL)
                  gtk_list_select_child(GTK_LIST(input->select->menu),
                                        input->select->options[i].menuitem);
            } else {
               if (input->select->options[i].menuitem->state == GTK_STATE_SELECTED)
                  gtk_list_unselect_child(GTK_LIST(input->select->menu),
                                          input->select->options[i].menuitem);
            }
         }
      }
      break;
      /* textarea */
   default:
      break;
   }
}


/*
 * Add a new input to the form data structure, setting the initial
 * values.
 */
static void Html_add_input(DilloHtmlForm *form,
                           DilloHtmlInputType type,
                           GtkWidget *widget,
                           const char *name,
                           const char *init_str,
                           gboolean init_val)
{
   DilloHtmlInput *input;

   a_List_add(form->inputs, form->num_inputs, sizeof(*form->inputs),
              form->num_inputs_max);
   input = &(form->inputs[form->num_inputs]);
   input->type = type;
   input->widget = widget;
   if (name != NULL)
      input->name = g_strdup(name);
   else
      input->name = NULL;
   if (init_str != NULL)
      input->init_str = g_strdup(init_str);
   else
      input->init_str = NULL;
   input->select = NULL;
   input->init_val = init_val;
   Html_reset_input(input);
   form->num_inputs++;
}


/*
 * Given a GtkWidget, find the form that contains it.
 * Return value: form_index if successful, -1 otherwise.
 */
static int Html_find_form(GtkWidget *reset, DilloHtmlLB *html_lb)
{
   gint form_index;
   gint input_index;
   DilloHtmlForm *form;

   for (form_index = 0; form_index < html_lb->num_forms; form_index++) {
      form = &(html_lb->forms[form_index]);
      for (input_index = 0; input_index < form->num_inputs; input_index++) {
         if (form->inputs[input_index].widget == reset) {
            return form_index;
         }
      }
   }
   return -1;
}

/*
 * Reset all inputs in the form containing reset to their initial values.
 * In general, reset is the reset button for the form.
 */
static void Html_reset_form(GtkWidget *reset, DilloHtmlLB *html_lb)
{
   gint i, j;
   DilloHtmlForm *form;

   if ( (i = Html_find_form(reset, html_lb)) != -1 ){
      form = &html_lb->forms[i];
      for ( j = 0; j < form->num_inputs; j++)
         Html_reset_input(&(form->inputs[j]));
   }
}

/*
 * Urlencode 'val' and append it to 'str'
 * -RL :: According to the RFC 1738, only alphanumerics, the special 
 *        characters "$-_.+!*'(),", and reserved characters ";/?:@=&" used 
 *        for their *reserved purposes* may be used unencoded within a URL.
 */
static void Html_urlencode_append(GString *str, const char *val)
{
   gint i;
   static const char *special = "$-_.+!*'(),";
   static const char *hex = "0123456789ABCDEF";

   if ( val == NULL )
      return;
   for (i = 0; val[i] != '\0'; i++) {
      if (val[i] == ' ') {
         g_string_append_c(str, '+');
      } else if (isalnum (val[i]) || strchr(special, val[i])) {
         g_string_append_c(str, val[i]);
      } else {
         g_string_append_c(str, '%');
         g_string_append_c(str, hex[(val[i] >> 4) & 15]);
         g_string_append_c(str, hex[val[i] & 15]);
      }
   }
}

/*
 * Append a name-value pair to an existing url.
 * (name and value are urlencoded before appending them)
 */
void Html_append_input(GString *url, const char *name, const char *value)
{
   Html_urlencode_append(url, name);
   g_string_append_c(url, '=');
   Html_urlencode_append(url, value);
   g_string_append_c(url, '&');
}

/*
 * Submit the form containing the submit input by making a new query URL
 * and sending it with a_Nav_push.
 */
static void Html_submit_form(GtkWidget *submit, DilloHtmlLB *html_lb)
{
   gint i, input_index;
   DilloHtmlForm *form;
   DilloHtmlInput *input;

   /* Search the form that generated the submition event */
   if ( (i = Html_find_form(submit, html_lb)) == -1 )
      return;

   /* Increment reference use on html_lb */
   html_lb->NRef++;

   form = &html_lb->forms[i];
   if ((form->method == DILLO_HTML_METHOD_GET) || 
       (form->method == DILLO_HTML_METHOD_POST)) {
      GString *new_url = g_string_new("");
      g_print("Html_submit_form form->action=%s\n", form->action);
      g_string_append(new_url, form->action);
      g_string_append_c(new_url, '?');
      if (form->method == DILLO_HTML_METHOD_POST)
         g_string_append(new_url, "!POST:");

      for (input_index = 0; input_index < form->num_inputs; input_index++) {
         input = &(form->inputs[input_index]);
         switch (input->type) {
           case DILLO_HTML_INPUT_TEXT:
           case DILLO_HTML_INPUT_PASSWORD:
              Html_append_input(new_url, input->name,
                                gtk_entry_get_text(GTK_ENTRY(input->widget)));
              break;
           case DILLO_HTML_INPUT_CHECKBOX:
           case DILLO_HTML_INPUT_RADIO:
              if (GTK_TOGGLE_BUTTON(input->widget)->active &&
                  input->name != NULL && input->init_str != NULL) {
                 Html_append_input(new_url, input->name, input->init_str);
              }
              break;
           case DILLO_HTML_INPUT_HIDDEN:
              Html_append_input(new_url, input->name, input->init_str);
              break;
           case DILLO_HTML_INPUT_SELECT:
              for (i = 0; i < input->select->num_options; i++) {
                 if (GTK_CHECK_MENU_ITEM(input->select->options[i].menuitem)->
                     active) {
                    Html_append_input(new_url, input->name,
                                      input->select->options[i].value);
                    break;
                 }
              }
              break;
           case DILLO_HTML_INPUT_SEL_LIST:
              for (i = 0; i < input->select->num_options; i++) {
                 if (input->select->options[i].menuitem->state ==
                     GTK_STATE_SELECTED) {
                    Html_append_input(new_url, input->name,
                                      input->select->options[i].value);
                 }
              }
              break;
              /* textarea */
           default:
              break;
         } /* switch */
      } /* for (inputs) */

      if ( new_url->str[new_url->len - 1] == '&' )
         g_string_truncate(new_url, new_url->len - 1);
      a_Nav_push(html_lb->bw, new_url->str);
      g_string_free(new_url, TRUE);
   } else {
      g_print("Method unknown\n");
   }

   /* Decrement use on html_lb */
   Html_lb_free(html_lb);
}

/*
 * Submit form if it has no submit button.
 * (Called by GTK+ when the user presses enter in a text entry within a form)
 */
static void Html_enter_submit_form(GtkWidget *submit, DilloHtmlLB *html_lb)
{
   gint i;

   /* Search the form that generated the submition event */
   if ( (i = Html_find_form(submit, html_lb)) == -1 )
      return;

   if ( html_lb->forms[i].HasSubmitButton == FALSE )
      Html_submit_form(submit, html_lb);
}

/*
 * Add a new input to current form
 */
static void Html_tag_open_input(DilloHtml *html, char *tag, gint tagsize)
{
   DilloHtmlForm *form;
   GtkWidget *widget;
   GtkWidget *event_box;
   char type[80];
   char value[1024];
   char name_str[1024];
   char checked[80];
   char *name, *init_str;
   gboolean init_val;
   gint input_index;
   GSList *group;
   DilloHtmlLB *html_lb;

   html_lb = html->linkblock;

   form = &(html_lb->forms[html_lb->num_forms - 1]);
   widget = NULL;
   if (!Html_get_attr(tag, tagsize, "type", type, sizeof(type)))
      strcpy(type, "text");

   if (Html_get_attr(tag, tagsize, "name", name_str, sizeof(name_str)))
      name = name_str;
   else
      name = NULL;
   /* move value param from case analysis to here? */
   if (!g_strcasecmp(type, "text")) {
      widget = gtk_entry_new();
      if (Html_get_attr(tag, tagsize, "value", value, sizeof(value)))
         init_str = value;
      else
         init_str = "";
      Html_add_input(form, DILLO_HTML_INPUT_TEXT,
                     widget, name, init_str, FALSE);
      gtk_signal_connect(GTK_OBJECT(widget), "activate",
                         GTK_SIGNAL_FUNC(Html_enter_submit_form),
                         html_lb);
   } else if (!g_strcasecmp(type, "password")) {
      widget = gtk_entry_new();
      gtk_entry_set_visibility(GTK_ENTRY(widget), FALSE);
      if (Html_get_attr(tag, tagsize, "value", value, sizeof(value)))
         init_str = value;
      else
         init_str = "";
      Html_add_input(form, DILLO_HTML_INPUT_PASSWORD,
                     widget, name, init_str, FALSE);
   } else if (!g_strcasecmp(type, "checkbox")) {
      event_box = gtk_event_box_new();
      widget = gtk_check_button_new();
      gtk_container_add(GTK_CONTAINER(event_box), widget);
      if (Html_get_attr(tag, tagsize, "value", value, sizeof(value)))
         init_str = value;
      else
         init_str = "";
      init_val = Html_get_attr(tag, tagsize, "checked", checked,
                                sizeof(checked));
      Html_add_input(form, DILLO_HTML_INPUT_CHECKBOX,
                     widget, name, init_str, init_val);
      gtk_widget_show(widget);
      widget = event_box;
   } else if (!g_strcasecmp(type, "radio")) {
      group = NULL;
      for (input_index = 0; input_index < form->num_inputs; input_index++) {
         if (form->inputs[input_index].type == DILLO_HTML_INPUT_RADIO &&
            (form->inputs[input_index].name && 
             !strcmp(form->inputs[input_index].name, name)) ){
            group = gtk_radio_button_group(GTK_RADIO_BUTTON
                                     (form->inputs[input_index].widget));
            break;
         }
      }
      event_box = gtk_event_box_new();
      widget = gtk_radio_button_new(group);
      gtk_container_add(GTK_CONTAINER(event_box), widget);
      if (Html_get_attr(tag, tagsize, "value", value, sizeof(value)))
         init_str = value;
      else
         init_str = "";
      init_val = Html_get_attr(tag, tagsize, "checked", checked,
                                sizeof(checked));
      Html_add_input(form, DILLO_HTML_INPUT_RADIO,
                     widget, name, init_str, init_val);
      gtk_widget_show(widget);
      widget = event_box;
   } else if (!g_strcasecmp(type, "hidden")) {
      if (Html_get_attr(tag, tagsize, "value", value, sizeof(value)))
         init_str = value;
      else
         init_str = "";
      Html_add_input(form, DILLO_HTML_INPUT_HIDDEN,
                     NULL, name, init_str, FALSE);
   } else if (!g_strcasecmp(type, "submit")) {
      if (Html_get_attr(tag, tagsize, "value", value, sizeof(value)))
         init_str = value;
      else
         init_str = "";

      if(init_str == "") {
         widget = gtk_button_new_with_label("submit");
      } else {
         widget = gtk_button_new_with_label(init_str);
      }
      form->HasSubmitButton = TRUE;
      Html_add_input(form, DILLO_HTML_INPUT_SUBMIT,
                     widget, name, init_str, FALSE);
      gtk_signal_connect(GTK_OBJECT(widget), "clicked",
                         GTK_SIGNAL_FUNC(Html_submit_form),
                         html_lb);
   } else if (!g_strcasecmp(type, "reset")) {
      if (Html_get_attr(tag, tagsize, "value", value, sizeof(value)))
         widget = gtk_button_new_with_label(value);
      else
         widget = gtk_button_new_with_label("Reset");
      Html_add_input(form, DILLO_HTML_INPUT_RESET,
                     widget, NULL, NULL, FALSE);
      gtk_signal_connect(GTK_OBJECT(widget), "clicked",
                         GTK_SIGNAL_FUNC(Html_reset_form),
                         html_lb);
   }
   if (widget != NULL) {
      gtk_widget_show(widget);
      a_Dw_page_add_widget((DwPage *) html->dw,
                           a_Dw_embed_gtk_new(widget),
                           html->stack[html->stack_top].attr);
   }
}

/*
 * ?
 */
/* The select tag is quite tricky, because of gorpy html syntax. */
static void Html_tag_open_select(DilloHtml *html, char *tag, gint tagsize)
{
   DilloHtmlForm *form;
   char name_str[1024];
   char size_str[80];
   char multiple[80];
   char *name;
   gint size;
   GtkWidget *widget;
   GtkWidget *menu;
   DilloHtmlSelect *Select;
   DilloHtmlLB *html_lb;

   html_lb = html->linkblock;

   form = &(html_lb->forms[html_lb->num_forms - 1]);
   if (Html_get_attr(tag, tagsize, "name", name_str, sizeof(name_str)))
      name = name_str;
   else
      name = NULL;
   if (Html_get_attr(tag, tagsize, "size", size_str, sizeof(size_str)))
      size = atoi(size_str);
   else
      size = 1;
   if (size < 1)
      size = 1;

   if (size == 1) {
      menu = gtk_menu_new();
      widget = gtk_option_menu_new();

      Html_add_input(form, DILLO_HTML_INPUT_SELECT,
                     widget, name, NULL, FALSE);
   } else {
      menu = gtk_list_new();

      Html_add_input(form, DILLO_HTML_INPUT_SEL_LIST,
                     menu, name, NULL, FALSE);
      if (Html_get_attr(tag, tagsize, "multiple", multiple,
                         sizeof(multiple)))
         gtk_list_set_selection_mode(GTK_LIST(menu), GTK_SELECTION_MULTIPLE);
   }

   Select = g_new(DilloHtmlSelect, 1);
   Select->menu = menu;
   Select->size = size;
   Select->num_options = 0;
   Select->num_options_max = 16;
   Select->options = g_new(DilloHtmlOption, Select->num_options_max);

   /* for consistency, the select should probably be another argument
    * to Html_add_input, but I felt bad about the number of
    * arguments that it had already, so I just set it here. */
   form->inputs[form->num_inputs - 1].select = Select;

   Html_stash_init(html);
}

/*
 * ?
 */
static void Html_option_finish(DilloHtml *html)
{
   DilloHtmlForm *form;
   DilloHtmlInput *input;
   GtkWidget *menuitem;
   GSList *group;
   DilloHtmlLB *html_lb;

   html_lb = html->linkblock;

   form = &(html_lb->forms[html_lb->num_forms - 1]);
   input = &(form->inputs[form->num_inputs - 1]);
   /* maybe break into two nested if's (num_options test on outside? */
   if (input->type == DILLO_HTML_INPUT_SELECT &&
       input->select->num_options > 0) {
      if (input->select->num_options == 1)
         group = NULL;
      else
         group = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM
                                   (input->select->options[0].menuitem));
      menuitem = gtk_radio_menu_item_new_with_label(group, html->stash_buf);
      input->select->options[input->select->num_options - 1].menuitem = menuitem;
      if (input->select->options[input->select->num_options - 1].value == NULL)
         input->select->options[input->select->num_options - 1].value =
             g_strdup(html->stash_buf);
      gtk_menu_append(GTK_MENU(input->select->menu), menuitem);
      if (input->select->options[input->select->num_options - 1].init_val)
         gtk_menu_item_activate(GTK_MENU_ITEM(menuitem));
      gtk_widget_show(menuitem);
   } else if (input->type == DILLO_HTML_INPUT_SEL_LIST &&
              input->select->num_options > 0) {
      menuitem = gtk_list_item_new_with_label(html->stash_buf);
      input->select->options[input->select->num_options - 1].menuitem = menuitem;
      if (input->select->options[input->select->num_options - 1].value == NULL)
         input->select->options[input->select->num_options - 1].value =
             g_strdup(html->stash_buf);
      gtk_container_add(GTK_CONTAINER(input->select->menu), menuitem);
      if (input->select->options[input->select->num_options - 1].init_val)
         gtk_list_select_child(GTK_LIST(input->select->menu), menuitem);
      gtk_widget_show(menuitem);
   }
}

/*
 * ?
 */
static void Html_tag_open_option(DilloHtml *html, char *tag, gint tagsize)
{
   DilloHtmlForm *form;
   DilloHtmlInput *input;
   char value[1024];
   char selected[80];
   DilloHtmlLB *html_lb;
   gint no;

   html_lb = html->linkblock;

   form = &(html_lb->forms[html_lb->num_forms - 1]);
   input = &(form->inputs[form->num_inputs - 1]);
   if (input->type == DILLO_HTML_INPUT_SELECT ||
       input->type == DILLO_HTML_INPUT_SEL_LIST) {
      Html_option_finish(html);
      no = input->select->num_options;
      a_List_add(input->select->options, no, sizeof(DilloHtmlOption),
                 input->select->num_options_max);
      input->select->options[no].menuitem = NULL;
      if (Html_get_attr(tag, tagsize, "value", value, sizeof(value)))
         input->select->options[no].value = g_strdup(value);
      else
         input->select->options[no].value = NULL;
      input->select->options[no].init_val =
          Html_get_attr(tag, tagsize, "selected", selected,
                         sizeof(selected));
      input->select->num_options++;
   }
   Html_stash_init(html);
}

/*
 * ?
 */
static void Html_tag_close_select(DilloHtml *html, char *tag, gint tagsize)
{
   DilloHtmlForm *form;
   DilloHtmlInput *input;
   GtkWidget *scrolledwindow;
   DilloHtmlLB *html_lb;

   html_lb = html->linkblock;

   form = &(html_lb->forms[html_lb->num_forms - 1]);
   input = &(form->inputs[form->num_inputs - 1]);
   if (input->type == DILLO_HTML_INPUT_SELECT) {
      Html_option_finish(html);

      gtk_option_menu_set_menu(GTK_OPTION_MENU(input->widget),
                               input->select->menu);
      Html_select_set_history(input);
#if 0
      gtk_option_menu_set_history(GTK_OPTION_MENU(input->widget), 1);
#endif

      gtk_widget_show(input->widget);
      a_Dw_page_add_widget((DwPage *) html->dw,
                           a_Dw_embed_gtk_new(input->widget),
                           html->stack[html->stack_top].attr);
   } else if (input->type == DILLO_HTML_INPUT_SEL_LIST) {
      Html_option_finish(html);

      if (input->select->size < input->select->num_options) {
         scrolledwindow = gtk_scrolled_window_new(NULL, NULL);
         gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwindow),
                                        GTK_POLICY_NEVER,
                                        GTK_POLICY_AUTOMATIC);
         gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW
                                               (scrolledwindow), 
                                               input->widget);
         
         /* Calculate the height of the scrolled window */
         gtk_widget_size_request(input->widget, &input->widget->requisition);
         gtk_widget_set_usize(scrolledwindow, -1, input->select->size *
                              (input->select->options->menuitem->
                               requisition.height));
         gtk_widget_show(input->widget);
         input->widget = scrolledwindow;
      }
      gtk_widget_show(input->widget);
      
      /* note: In this next call, scrolledwindows get a g_warning from
       * gdkwindow.c:422. I'm not really going to sweat it now - the
       * embedded widget stuff is going to get massively redone anyway. */
      a_Dw_page_add_widget((DwPage *) html->dw,
                           a_Dw_embed_gtk_new(input->widget),
                           html->stack[html->stack_top].attr);
   }
   Html_pop_tag(html, tag, tagsize);
}

/*
 * Default close for most tags - just pop the stack.
 */
static void Html_tag_default_close(DilloHtml *html, char *tag, gint tagsize)
{
   Html_pop_tag(html, tag, tagsize);
}

/*
 * Default close for paragraph tags - pop the stack and break.
 */
static void Html_tag_par_close(DilloHtml *html, char *tag, gint tagsize)
{
   Html_pop_tag(html, tag, tagsize);
   a_Dw_page_parbreak((DwPage *) html->dw, 9);
}

/*
 * Default close for tags with no close (e.g. <br>) - do nothing
 */
static void Html_tag_nop_close(DilloHtml *html, char *tag, gint tagsize)
{
}


/* HTML tag-handling function list */
static TagFunct F_list[] = {
   Html_tag_par_close,       /* 0 */
   Html_tag_nop_close,
   Html_tag_default_close,
   Html_tag_open_title, 
   Html_tag_close_title,
   Html_tag_open_body,       /* 5 */
   Html_tag_open_p,
   Html_tag_open_h,
   Html_tag_open_br,
   Html_tag_open_b,
   Html_tag_open_strong,     /* 10 */
   Html_tag_open_i,
   Html_tag_open_em,
   Html_tag_open_cite,
   Html_tag_open_tt,
   Html_tag_open_button,     /* 15 */
   Html_tag_open_img,
   Html_tag_open_a,
   Html_tag_open_blockquote,
   Html_tag_open_ul,
   Html_tag_open_ol,         /* 20 */
   Html_tag_open_li,
   Html_tag_open_hr,
   Html_tag_open_dl,
   Html_tag_open_dt,
   Html_tag_open_dd,         /* 25 */
   Html_tag_open_pre,
   Html_tag_open_form,
   Html_tag_open_input,
   Html_tag_open_option,
   Html_tag_open_select,     /* 30 */
   Html_tag_close_select,
   Html_tag_open_td
};

typedef struct _TagInfo {
   gchar *name;
   guchar open, close, tagsize;
} TagInfo;

/*
 * Function index for the open and close functions for each tag 
 * (Alphabetically sorted for a binary search)
 */
#define NTAGS 35
static TagInfo Tags[NTAGS] = {
   {"a", 17,  2},            /* 0 */
   {"address", 11,  0},
   {"b",  9,  2},
   {"blockquote", 18,  0},
   {"body",  5,  2},
   {"br",  8,  1},           /* 5 */
   {"button", 15,  1},
   {"cite", 13,  2},
   {"dd", 25,  0},
   {"dl", 23,  0},
   {"dt", 24,  0},           /* 10 */
   {"em", 12,  2},
   {"form", 27,  2},
   {"h1",  7,  0},
   {"h2",  7,  0},
   {"h3",  7,  0},           /* 15 */
   {"h4",  7,  0},
   {"h5",  7,  0},
   {"h6",  7,  0},
   {"hr", 22,  1},
   {"i", 11,  2},            /* 20 */
   {"img", 16,  1},
   {"input", 28,  1},
   {"li", 21,  2},
   {"ol", 20,  0},
   {"option", 29,  1},       /* 25 */
   {"p",  6,  0},
   {"pre", 26,  0},
   {"select", 30, 31},
   {"strong", 10,  2},
   {"td", 32,  2},           /* 30 */   /* td is disabled */
   {"title",  3,  4},
   {"tr",  8,  1},
   {"tt", 14,  2},
   {"ul", 19,  0}
};


/*
 * Compares tag from buffer ('/' or '>' or ' ' ended string) [p1]
 * with tag from taglist (lowercase, zero ended string) [p2]
 * Return value: as strcmp()
 */
gint Html_tag_compare(char *p1, char *p2)
{
   while ( *p2 ) {
      if ( tolower(*p1) != *p2 )
         return(tolower(*p1) - *p2);
      ++p1; 
      ++p2;
   }
   return !strchr("/> \n", *p1);
}

/*
 * Get 'tag' index
 * return -1 if tag is not handled yet
 */
gint Html_tag_index(char *tag)
{
   gint low, high, mid, cond;

   /* Binary search */
   low = 0;
   high = NTAGS - 1;          /* Last tag index */
   while (low <= high) {
      mid = (low + high) / 2;
      if ((cond = Html_tag_compare(tag, Tags[mid].name)) < 0 )
         high = mid - 1;
      else if (cond > 0)
         low = mid + 1;
      else
         return mid;
   }
   return -1;
}

/*
 * Process a tag, given as 'tag' and 'tagsize'.
 * ('tag' must include the enclosing angle brackets)
 * This function calls the right open or close function for the tag.
 */
static void Html_process_tag(DilloHtml *html, char *tag, gint tagsize)
{
   gint i;
   char *start;

   /* discard the '<' */
   start = tag + 1;
   /* discard junk */
   while ( strchr(" \n", *start) )
      ++start;

   i = Html_tag_index(start + (*start == '/'));
   if (i != -1) {
      if (*start != '/') {
         /* Open function */
         (*F_list[Tags[i].open]) (html, tag, tagsize);
      } else {
         /* Close function */
         (*F_list[Tags[i].close]) (html, tag, tagsize);
      }
   } else {
      /* tag not working - just ignore it */
   }
}

/*
 * Match the tag in the tag, tagsize structure with 'tagstr'.
 * 'tag' has the initial < skipped over (e.g. "P align=center>")
 * Return value:
 *    strlen(tagstr)   if found
 *    0                if not
 */
static gint Html_match_attr(const char *tagstr, const char *tag, gint tagsize)
{
   gint i;

   for (i = 0; i < tagsize && tagstr[i] != '\0'; i++) {
      if (tolower(tagstr[i]) != tolower(tag[i]))
         return 0;
   }
   if (i < tagsize && (isspace(tag[i]) || tag[i] == '>' || tag[i] == '='))
      return i;
   return 0;
}

/*
 * If pointing to a parameter value, copy the value into 'data'
 * Return value:
 *   1 on success
 *   0 if value is bigger than datasize
 * (This code is complex, but avoids a second pass over the value  --Jcid)
 */
static gint Html_get_attr_value(const char *tag, gint tagsize,
                                char *data, gint datasize)
{
   gint i;

   if (tag[0] == '"' || tag[0] == '\'' ) {
      /* copy to end quote or to datasize */
      for (i = 1; i < tagsize && i < datasize && 
           (*data++ = tag[i]) != tag[0]; i++)
         ;
   } else {
      for (i = 0; i < tagsize && i < datasize &&
           !isspace((*data++ = tag[i])) && tag[i] != '>'; i++)
         ;
      /* xml support: allow tag to be closed by "/>" */
      if (i == tagsize - 1 && i >= 1 && tag[i - 1] == '/')
         i--;
   }

   if ( i < tagsize && i < datasize ) {
      *--data = '\0';
      return 1;
   }

   return 0;
}

/*
 * Get attribute value for 'attrname'
 *  Tags start with '<' and end with a '>' (Ex: "<P align=center>")
 *  tagsize = strlen of the tag from '<' to '>', inclusive.
 *
 * Return 0 if param is not present, 1 if present, -1 if datasize is too
 * small. If nonzero, param is stored into data.
 */ 
static gint Html_get_attr(const char *tag, gint tagsize,
                           const char *attrname,
                           char *data, gint datasize)
{
   gint i, len;

   i = 1;
   while ( i < tagsize ) {
      /* Search token start */
      for ( ; i < tagsize && !isspace(tag[i]); i++);
      for ( ; i < tagsize && isspace(tag[i]); i++);

      if ( (len = Html_match_attr(attrname, tag + i, tagsize - i)) != 0 ){
         /* Found the attribute name */
         for (i += len; i < tagsize && isspace(tag[i]); i++);
         if ( i < tagsize && tag[i] == '=' ){
            i++;
            if ( !Html_get_attr_value(tag + i, tagsize - i, data, datasize) )
               return -1;
         } else {
            *data = '\0';
         }
         return 1;
      }
   }
   return 0;
}

/*
 * Dispatch the apropriate function for 'Op'
 * This function is a Cache client and gets called whenever new data arrives
 *  Op      : operation to perform.
 *  CbData  : a pointer to a DilloHtml structure
 *  Buf     : a pointer to new data
 *  BufSize : new data size (in bytes)
 */
static void Html_callback(int Op, void *CbData, guchar *Buf, gint BufSize)
{
   if (Op)
      Html_close(CbData);
   else
      Html_write(CbData, Buf, BufSize);
}

/*
 * Here's where we parse the html and put it into the page structure.
 * (This function is called by Html_callback whenever there's new data)
 */
static void Html_write(DilloHtml *html, char *Buf, gint BufSize)
{
   DwPage *page;
   char completestr[32];
   gint token_start, buf_index;
   char *buf = Buf + html->Start_Ofs;
   gint bufsize = BufSize - html->Start_Ofs;
   char ch;

   Html_page_check(html);
   page = (DwPage *) html->dw;
   a_Dw_page_update_begin(page);

   /* Here's where we parse the html and put it into the page structure. */

   /* Now, Buf and Bufsize define a buffer aligned to start at a token
    * boundary. Iterate through tokens until end of buffer is reached. */
   buf_index = 0;
   token_start = buf_index;
   while (buf_index < bufsize) {

#ifdef VERBOSE
      g_print("handling token %c%c%c\n", buf[buf_index], buf[buf_index + 1],
              buf[buf_index + 2]);
#endif

      /* invariant: buf_index == bufsize || token_start == buf_index */

      if (isspace(buf[buf_index])
          && (html->stack[html->stack_top].parse_mode
              != DILLO_HTML_PARSE_MODE_PRE)) {
         /* whitespace : consume all available whitespace */
         do {
            buf_index++;
         } while (buf_index < bufsize && isspace(buf[buf_index]));
         Html_process_space(html, buf + token_start,
                            buf_index - token_start);
         token_start = buf_index;
      } else if (buf[buf_index] == '<') {
         if (buf_index + 4 <= bufsize &&
             buf[buf_index + 1] == '!' &&
             buf[buf_index + 2] == '-' &&
             buf[buf_index + 3] == '-') {
            /* a comment: search for close of comment, skipping over
             * everything except a matching "-->" string. */
            buf_index += 6;
            if (buf_index > bufsize)
               buf_index = bufsize;

            while (buf_index < bufsize &&
                   (buf[buf_index] != '>' ||
                    buf[buf_index - 1] != '-' ||
                    buf[buf_index - 2] != '-'))
               buf_index++;

            if (buf_index < bufsize) {
               buf_index++;
               /* if you ever want to pass the comments downstream, put
                * code here. You'll need to do this for J*Script and
                * style sheets. */
               token_start = buf_index;
            }
         } else {
            /* a tag: search for close of tag (skipping over quoted strings) */
            buf_index++;
            while (buf_index < bufsize && (ch = buf[buf_index]) != '>') {
               if ( ch == '"' || ch == '\'' ) {
                  /* Skip over quoted string */
                  buf_index++;
                  while (buf_index < bufsize && buf[buf_index++] != ch);
               } else {
                  buf_index++;
               }
            }
            if (buf_index < bufsize) {
               buf_index++;
               Html_process_tag(html, buf + token_start,
                                buf_index - token_start);
               token_start = buf_index;
            }
         }
      } else {
         /* a word : search for whitespace or tag open */
         do {
            buf_index++;
         } while (buf_index < bufsize && !isspace(buf[buf_index]) &&
                  buf[buf_index] != '<');
         if (buf_index < bufsize) {
            /* successfully found end of token */
            Html_process_word(html, buf + token_start,
                              buf_index - token_start);
            token_start = buf_index;
         }
      }
   }

   html->Start_Ofs += token_start;

   if ( html->bw ) {
      sprintf(completestr,"Page: %.1f Kb", (float)html->Start_Ofs/1024);
      a_Progressbar_update(html->bw->progress, completestr, 1);
   }

   a_Dw_page_update_end(page);
}

/*
 * ?
 */
static void Html_close(DilloHtml *html)
{
   gint i;

   if (((DwPage *) html->dw)->Parent) {
      ((DwPage *) html->dw)->Parent = NULL;
   }
   for (i = 0; i <= html->stack_top; i++)
      g_free(html->stack[i].tag);
   g_free(html->stack);

   g_free(html->stash_buf);

   /* Let the user know we finished loading the page */
   if ( html->bw )
      a_Progressbar_update(html->bw->progress, "", 0);

   g_free(html);
}
